﻿using Caliburn.Micro;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Media3D;
using VisualComponents.Create3D;
using VisualComponents.UX;
using VisualComponents.UX.Input;
using VisualComponents.UX.Viewport;
using Color = System.Windows.Media.Color;
using IManipulator = VisualComponents.Create3D.IManipulator;


public interface IWpfGraphics
{
    event System.Action CameraMoved;
    void Clear();
    void Clear(bool everything);
    void Add(GeometryModel3D model);
    void Add(params Model3D[] models);
    void Add(params GeometryModel3D[] models);
    void Add(MeshGeometry3D mesh, Color color);
}

/// <summary>
/// Class for rendering wpf geometries.
/// </summary>
public class WpfGraphics : IWpfGraphics
{
    public event System.Action CameraMoved;
    public List<GeometryModel3D> GeometryModel3Ds = new List<GeometryModel3D>();
    public List<Model3D> Models3Ds = new List<Model3D>();
    public IGeometryHolder GeomHolder;
    public bool IsAdded;

    public void Clear()
    {
        if (GeomHolder == null)
            GeomHolder = IoC.Get<IGeometryHolder>();
        GeomHolder.Clear(this, true);
    }
    public void Clear(bool everything)
    {
        if (GeomHolder == null)
            GeomHolder = IoC.Get<IGeometryHolder>();
        GeomHolder.Clear(this, everything);
    }

    public void Add(Model3D model)
    {
        if (GeomHolder == null)
            GeomHolder = IoC.Get<IGeometryHolder>();
        GeomHolder.Add(this, model);
    }

    public void Add(GeometryModel3D model)
    {
        if (GeomHolder == null)
            GeomHolder = IoC.Get<IGeometryHolder>();
        GeomHolder.Add(this, model);
    }

    public void Add(params Model3D[] models)
    { 
        foreach (var model in models)
        {
            Add(model);
        }
    }

    public void Add(params GeometryModel3D[] models)
    {
        foreach (var model in models)
        {
            Add(model);
        }
    }
    
    public void Add(MeshGeometry3D mesh, Color color)
    {
        if (GeomHolder == null)
            GeomHolder = IoC.Get<IGeometryHolder>();
        GeomHolder.Add(this, mesh, color);
    }
    public void OnCameraMoved()
    {
        if (CameraMoved != null)
        {
            CameraMoved();
        }
    }
    public void ClearCameraMoved()
    {
        CameraMoved = null;
    }
}

public interface IGeometryHolder
{
    void Clear(WpfGraphics wpfGraphics, bool everything);
    void Add(WpfGraphics wpfGraphics, GeometryModel3D model);
    void Add(WpfGraphics wpfGraphics, MeshGeometry3D mesh, Color color);
    void Add(WpfGraphics wpfGraphics, Model3D mesh);
}

[Export(typeof(IGeometryHolder))]
[Export(typeof(IManipulator))]
[PartCreationPolicy(CreationPolicy.Shared)]
public class GeometryHolder : ManipulatorBase, IGeometryHolder
{
    public ModelVisual3D _visualChildren;
    public Model3DGroup _visualChildrenGroup;
    private List<WpfGraphics> wpfGraphicsList = new List<WpfGraphics>();
    private bool listeningToCamera;

    public GeometryHolder()
    {
    }

    public override void OnCreate()
    {
        _visualChildren = new ModelVisual3D();
        _visualChildrenGroup = new Model3DGroup();
        _visualChildren.Content = _visualChildrenGroup;
        Children.Add(_visualChildren);
        RenderOptions.SetBitmapScalingMode(_visualChildren, BitmapScalingMode.NearestNeighbor);
    }

    public void OnCameraChanged(object sender, ICamera camera)
    {
        var prev = wpfGraphicsList.ToArray();
        foreach (var wpfGraphics in prev)
        {
            wpfGraphics.OnCameraMoved();
        }
    }

    public void Clear(WpfGraphics wpfGraphics, bool everything)
    {
        if (everything)
        {
            wpfGraphics.ClearCameraMoved();
            if (wpfGraphics.IsAdded)
            {
                wpfGraphics.IsAdded = false;
                wpfGraphicsList.Remove(wpfGraphics);
            }
        }
        if (wpfGraphics.GeometryModel3Ds != null)
        {
            if (wpfGraphics.GeometryModel3Ds.Count > 0)
            {
                foreach (var model in wpfGraphics.GeometryModel3Ds)
                {
                    _visualChildrenGroup.Children.Remove(model);
                }
                wpfGraphics.GeometryModel3Ds.Clear();
            }
        }

        if (wpfGraphics.Models3Ds != null)
        {
            if (wpfGraphics.Models3Ds.Count > 0)
            {
                foreach (var model in wpfGraphics.Models3Ds)
                {
                    _visualChildrenGroup.Children.Remove(model);
                }
            }
        }
        
    }

    public void Add(WpfGraphics wpfGraphics, GeometryModel3D model)
    {
        EnsureAdded(wpfGraphics);
        wpfGraphics.GeometryModel3Ds.Add(model);
        _visualChildrenGroup.Children.Add(model);
        RenderOptions.SetBitmapScalingMode(model, BitmapScalingMode.NearestNeighbor);
    }

    public void Add(WpfGraphics wpfGraphics, Model3D model)
    {
        EnsureAdded(wpfGraphics);
        wpfGraphics.Models3Ds.Add(model);
        _visualChildrenGroup.Children.Add(model);
        RenderOptions.SetBitmapScalingMode(model, BitmapScalingMode.NearestNeighbor);
    }

    public void Add(WpfGraphics wpfGraphics, MeshGeometry3D mesh, Color color)
    {
        EnsureAdded(wpfGraphics);
        var model = new GeometryModel3D();
        model.Geometry = mesh;
        model.Material = GetMaterial(color);
        model.BackMaterial = model.Material;
        wpfGraphics.GeometryModel3Ds.Add(model);
        _visualChildrenGroup.Children.Add(model);
        RenderOptions.SetBitmapScalingMode(model, BitmapScalingMode.NearestNeighbor);
    }

    private void EnsureAdded(WpfGraphics wpfGraphics)
    {
        if (!listeningToCamera)
        {
            listeningToCamera = true;
            var viewport = (Screen)IoC.Get<IViewport>();
            var viewport_helper = (viewport.GetView() as FrameworkElement).FindChild<HelperViewport>();
            viewport_helper.CameraChanged += OnCameraChanged;
        }
        if (!wpfGraphics.IsAdded)
        {
            wpfGraphics.IsAdded = true;
            wpfGraphicsList.Add(wpfGraphics);
        }
    }

    private static MaterialGroup GetMaterial(Color color)
    {
        MaterialGroup material = new MaterialGroup();
        SolidColorBrush brush = new SolidColorBrush(color);
        material.Children.Add(new DiffuseMaterial(brush));
        material.Children.Add(new EmissiveMaterial(brush));
        return material;
    }

    public override bool OnShow()
    {
        return true;
    }

    public override void OnPosition()
    {

    }

}
