﻿using Caliburn.Micro;
using HelixToolkit.Wpf;
using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Windows;
using System.Windows.Media.Media3D;
using VisualComponents.Create3D;
using VisualComponents.UX.Shared;
using Color = System.Windows.Media.Color;
using Matrix = VisualComponents.Create3D.Matrix;

namespace Plugin.SnapGraphicsAction
{
    [Export(typeof(IActionItem))]
    [Export(typeof(IPlugin))]
    public class SnapGraphics : ActionItem, IPlugin, IHandle<SnapResultMessage>, 
        IHandle<RequestActionAbortMessage>
    {
        SnapAction _snap;
        List<Matrix> _recorded_positions;
        SnapResult _last_result;
        IWpfGraphics _snappingGraphics = new WpfGraphics();
        IWpfGraphics _snapCompletedGraphics = new WpfGraphics();
        IEventAggregator _eventAggregator;
        IWindow3D _window;
        IApplication _app;
        List<ILayoutSchemaPropertyList> _Schemas;
        ILayoutSchemaPropertyList _schema;

        Vector3 _surfaceNormal;
        string _tooltipText = "Pick a point on edge";
        Vector3 _coneOrigin;

        private bool _canExecute;
        public new bool CanExecute
        {
            get
            {
                return _canExecute;
            }
            set
            {
                if (_canExecute != value)
                {
                    _canExecute = value;
                    NotifyOfPropertyChange(() => CanExecute);
                }
            }
        }

        private bool _isChecked;
        public bool IsChecked
        {
            get
            {
                return _isChecked;
            }
            set
            {
                if (value != _isChecked)
                {
                    _isChecked = value;
                    Start(_isChecked);
                    NotifyOfPropertyChange(() => IsChecked);
                }
            }
        }

        public SnapGraphics() : base("Snap3DGraphics", "Pick Edge", "", 
            null, MenuTools.ToggleButton, false, false)
        {

        }


        /// <summary>
        /// Start snapping
        /// </summary>
        /// <param name="value"></param>
        public void Start(bool value)
        {
            if (value)
            {
                _snap = new SnapAction(true, null, new SnapType[] {
                    SnapType.Edge,
                    SnapType.EdgeCurve,
                    SnapType.EdgeLine,
                    SnapType.EdgePoint });
                _snap.ContinuousMode = true;
                _snap.WantSurfaceNormal = true;
                _snap.SnapOnFullTopology = true;
                _snap.SnapOnSinglePart = true;
                _snap.ContinuousResultMouseDown += OnSnap;
                _snappingGraphics.Clear();
                _snappingGraphics.CameraMoved += OnCameraMoved;
                _snap.ContinuousResultMouseMove += _snap_ContinuousResultMouseMove;
                _snap.Completed += _snap_Completed;

                _recorded_positions = new List<Matrix>();
                //_Schemas = new List<ILayoutSchemaPropertyList>();
            }
            else
            {
                _snappingGraphics.CameraMoved -= OnCameraMoved;
                _snap.ContinuousResultMouseMove -= _snap_ContinuousResultMouseMove;
                _snappingGraphics.CameraMoved -= OnCameraMoved;

            }
        }

 
        private void _snap_Completed(object sender, ResultCompletionEventArgs e)
        {
            if (!e.WasCancelled)
            {
                _snapCompletedGraphics.Clear(true);
                GenerateOtherObjects();
            }
            if (e.WasCancelled)
            {
                _snapCompletedGraphics.Clear(true);
                _snappingGraphics.Clear(true);
                if (_recorded_positions != null)
                    _recorded_positions.Clear();

                //if (_Schemas.Count > 0)
                //{
                //    foreach (ILayoutSchemaPropertyList schema in _Schemas)
                //    {
                //        schema.Clear();
                //    }
                //}
            }

            IoC.Get<IRenderService>().RequestRender();
        }

        public void Handle(RequestActionAbortMessage message)
        {
            if (_snap != null)
            {
                IsChecked = false;
                _snapCompletedGraphics.Clear();
                _snapCompletedGraphics.Clear(true);
                _snappingGraphics.Clear();


                if (_recorded_positions != null)
                    _recorded_positions.Clear();
                //if (_Schemas.Count > 0)
                //{
                //    foreach (ILayoutSchemaPropertyList schema in _Schemas)
                //    {
                //        schema.Clear();
                //    }
                //}
            }
        }

        private void _snap_ContinuousResultMouseMove(object sender, ResultCompletionEventArgs e)
        {
            if (_snap.SnapResult.VisualObject != null)
            {
                if (_snap.SnapResult.SurfacePosition != null)
                {
                    _surfaceNormal = _snap.SnapResult.SurfacePosition.Value;
                    _surfaceNormal.Normalize();
                    _snappingGraphics.Clear();
                    VisualizeEdgeNormalsWith3DGraphics();
                }
            }
            else
            {
                _snappingGraphics.Clear();
            }
        }


        public void OnCameraMoved()
        {
            VisualizeEdgeNormalsWith3DGraphics();
        }

        private void VisualizeEdgeNormalsWith3DGraphics()
        {
            _snappingGraphics.Clear(false);

            var point = _snap.SnapResult.Position;

            if (_snap.SnapResult.SurfaceNormal != null)
            {
                _surfaceNormal = _snap.SnapResult.SurfaceNormal.Value;
                _surfaceNormal.Normalize();

                var direction = (Vector3D)_surfaceNormal;
                var origin = (Point3D)point;
                var coneBuilder = new MeshBuilder();
                double pixelScale = _app.ActiveWindow.GetProjectedPointPixelSize(point);
                double baseRadius = 0.1 * pixelScale;
                double topRadius = 4 * pixelScale;
                double height = 60 * pixelScale;
                coneBuilder.AddCone(origin, direction, baseRadius, topRadius, height, true, true, 8);
                _snappingGraphics.Add(coneBuilder.ToMesh(), Color.FromRgb(255, 255, 255));
            }
        }


        public void OnSnap(object sender, ResultCompletionEventArgs e)
        {
            if (_snap.SnapResult.HasResult)
            {
                if (_snap.SnapResult.SnapType == SnapType.Edge ||
                    _snap.SnapResult.SnapType == SnapType.EdgeCurve ||
                    _snap.SnapResult.SnapType == SnapType.EdgeLine ||
                    _snap.SnapResult.SnapType == SnapType.EdgePoint)
                {
                    _last_result = _snap.SnapResult;

                    var edge = _snap.SnapResult.GeoPrimitive as IEdge;                   

                    var endPoints = edge.Endpoints;

                    var obj = _last_result.VisualObject;

                    if (obj is ISimNode)
                    {
                        var snapedComponent = ((ISimNode)_last_result.VisualObject).Component;

                        var end1 = endPoints.Item1 + snapedComponent.TransformationInWorld.GetP();
                        var end2 = endPoints.Item2 + snapedComponent.TransformationInWorld.GetP();

                        var pointMatrix1 = new Matrix();
                        var pointMatrix2 = new Matrix();
                        pointMatrix1.SetP(end1);
                        pointMatrix2.SetP(end2);

                        pointMatrix1.SetWPR(snapedComponent.TransformationInWorld.GetWPR());
                        pointMatrix2.SetWPR(snapedComponent.TransformationInWorld.GetWPR());

                        List<Matrix> positions = new List<Matrix>();
                        positions.Add(pointMatrix1);
                        positions.Add(pointMatrix2);

                        foreach (Matrix position in positions)
                        {
                            Generate3DGraphicsAfterSnapping(position);
                            _recorded_positions.Add(position);
                        }
                    }
                }
                else
                {
                    _snappingGraphics.Clear();
                }
            }
            else
            {

                _snap.EndAction(false, true);
                IsChecked = false;
            }
        }

        private void SnapedComponent_TransformationChanged(object sender, EventArgs e)
        {
            
        }

        public void Generate3DGraphicsAfterSnapping(Matrix Position)
        {
            if (_last_result.VisualObject == null)
                return;
            var SnappedComponent = ((ISimNode)_last_result.VisualObject).Component;
            if (SnappedComponent == null)
                return;

            Point3D originTopCone = new Point3D();
            Point3D originSphere = new Point3D();
            
            if (_snap.SnapResult.SurfaceNormal != null)
            {
                var Point = Position.GetP();

                // The white lines are created with this
                CreateTrianglesAfterSnapping(Position);

                _surfaceNormal = _snap.SnapResult.SurfaceNormal.Value;
                _surfaceNormal.Normalize();

                var direction = (Vector3D)_surfaceNormal;

                originTopCone.X = Point.X;
                originTopCone.Y = Point.Y;
                originTopCone.Z = _coneOrigin.Z;

                originSphere.X = Point.X;
                originSphere.Y = Point.Y;
                originSphere.Z = Point.Z;

                double pixelScale = _app.ActiveWindow.GetProjectedPointPixelSize((Vector3)originSphere);
                double baseRadius = 4 * pixelScale;
                double topRadius = 0.1 * pixelScale;
                double height = 40 * pixelScale;

                var coneBuilder = new MeshBuilder();
                coneBuilder.AddCone(originTopCone, direction, topRadius, baseRadius, height, true, false, 8);
                _snapCompletedGraphics.Add(coneBuilder.ToMesh(), Color.FromRgb(255, 255, 255));

                var sphereBuilder = new MeshBuilder();
                double sphereRadius = 4 * pixelScale;
                sphereBuilder.AddSphere(originSphere, sphereRadius, 8);
                _snapCompletedGraphics.Add(sphereBuilder.ToMesh(), Color.FromRgb(255, 255, 255));

            }
            IoC.Get<IConsoleViewModel>().AppendMessage(Position.GetWPR().Y.ToString());
        }

       
        private void CreateTrianglesAfterSnapping(Matrix Position)
        {          
            Color color = Color.FromRgb(255, 255, 255);

            var Point = Position.GetP();
            Point.W = Position.GetWPR().W;
            double pixelScale = _app.ActiveWindow.GetProjectedPointPixelSize(Point);

            // This is the vector for the second triangle vertix
            var point2 = new Vector3();
            point2.X = Point.X + 0.8 * pixelScale;
            point2.Y = Point.Y + 0.8 * pixelScale;
            point2.Z = Point.Z + 200 * pixelScale;
            

            // Need this to set the starting point of cone in world
            _coneOrigin = point2;
            //_coneOrigin.Y -= 0.5;

            // This is the vector for the third triangle vertix
            Vector3 point3 = new Vector3();
            point3.X = Point.X + 0 * pixelScale;
            point3.Y = Point.Y + 0 * pixelScale;
            point3.Z = Point.Z + 200 * pixelScale;

            // Use MeshBuilder to generate first triangle
            MeshBuilder firstTriangle = new MeshBuilder();
            firstTriangle.AddTriangle((Point3D)Point, (Point3D)point2, (Point3D)point3);
            MeshGeometry3D firstTriangleModel = firstTriangle.ToMesh();

            // Add the first triangle to VC using the WpfGraphics helper class
            _snapCompletedGraphics.Add(firstTriangleModel, color);

            // Vector for second triangle
            Vector3 point5 = new Vector3();
            point5.X = Point.X + 0.8 * pixelScale;
            point5.Y = Point.Y + 0 * pixelScale;
            point5.Z = Point.Z + 0 * pixelScale;


            MeshBuilder secondTriangle = new MeshBuilder();
            secondTriangle.AddTriangle((Point3D)Point, (Point3D)point5, (Point3D)point2);
            MeshGeometry3D secondTriangleModel = secondTriangle.ToMesh();

            // Add the second triangle
            _snapCompletedGraphics.Add(secondTriangleModel, color);
        }


        /// <summary>
        /// Generate afer snap completion (i.e ILayoutSchemaPropertyList)
        /// </summary>
        public void GenerateOtherObjects()
        {

            if (_last_result.VisualObject == null)
                return;
            var snap_component = ((ISimNode)_last_result.VisualObject).Component;
            if (snap_component == null)
                return;

            
            // Generate 3D WPFs
            foreach (Matrix position in _recorded_positions)
            {
                Point3D originSphere = new Point3D();
                originSphere.X = position.GetP().X;
                originSphere.Y = position.GetP().Y;
                originSphere.Z = position.GetP().Z;
               
                double pixelScale = _app.ActiveWindow.GetProjectedPointPixelSize((Vector3)originSphere);
                var sphereBuilder = new MeshBuilder();
                double sphereRadius = 8 * pixelScale;
                sphereBuilder.AddSphere(originSphere, sphereRadius, 8);
                _snapCompletedGraphics.Add(sphereBuilder.ToMesh(), Color.FromRgb(0, 255, 0));
            }

            // OR Generate schema frames

            //_schema = _app.World.CreateLayoutItem<ILayoutSchemaPropertyList>("SnapSchema");
            //_Schemas.Add(_schema);

            //_schema.Node = snap_component.RootNode;

            //_schema.CreateSchemaProperty(typeof(Matrix), "Position");
            //_schema.CreateSchemaProperty(typeof(bool), "FrameVisibility");
            //_schema.CreateSchemaProperty(typeof(bool), "LineVisibility");
            //_schema.CreateSchemaProperty(typeof(Vector3), "LineColor");

            //_schema.FrameDataSchemaIndex = 0;
            //_schema.FrameVisibleSchemaIndex = 1;
            //_schema.LineDisplaySchemaIndex = 0;
            //_schema.LineVisibleSchemaIndex = 2;
            //_schema.LineColorSchemaIndex = 3;

            //Vector3 color = new Vector3(0, 1, 0);

            //_schema.Reserve(_recorded_positions.Count);

            //for (int i = 0; i < _recorded_positions.Count; i++)
            //{             
            //    _schema.SetSchemaValue(i, 0, _recorded_positions[i]);
            //    _schema.SetSchemaValue(i, 1, false);
            //    _schema.SetSchemaValue(i, 2, true);
            //    _schema.SetSchemaValue(i, 3, color);
            //}
            //_schema.IsHighlighted = true;

            //_schema.Update();
            //IoC.Get<IRenderService>().RequestRender();
        }


        /// <summary>
        /// Displays info on 3D world during snapping
        /// </summary>
        /// <param name="message"></param>
        /// <param name="description"></param>
        /// <param name="mousePosition"></param>
        private void PostMessage(string message, string description, Point mousePosition)
        {
            bool showMessage = !string.IsNullOrEmpty(message);

            ToolTipDisplayMessage ttMsg = new ToolTipDisplayMessage
            {
                Show = showMessage,
                Description = description,
                Title = message,
                X = mousePosition.X + 50,
                Y = mousePosition.Y + 50
            };

            _eventAggregator.Publish(ttMsg);
        }

        /// <summary>
        /// Snap result event
        /// </summary>
        /// <param name="message"></param>
        public void Handle(SnapResultMessage message)
        {
            PostMessage(_tooltipText, message.SnapResultText, message.MousePosition);
        }

        private void _app_LayoutClosed(object sender, LayoutClosedEventArgs e)
        {
            _snapCompletedGraphics.Clear(true);
            if (_recorded_positions!=null)
                _recorded_positions.Clear();
            if(_Schemas!= null)
                _Schemas.Clear();
            if (_schema!= null)
                _schema.Clear();

            IoC.Get<IRenderService>().RequestRender();
        }

        /// <summary>
        /// This is called when application starts
        /// </summary>
        public void Initialize()
        {
            vcAppStart();
        }

        private void vcAppStart()
        {
            registerActionItem();
            CanExecute = true;
            _eventAggregator = IoC.Get<IEventAggregator>();
            _eventAggregator.Subscribe(this);
            _window = IoC.Get<IApplication>().ActiveWindow;
            _app = IoC.Get<IApplication>();
            _app.LayoutClosed += _app_LayoutClosed;
        }
        public void Exit(){}

        /// <summary>
        /// Add Action to UI
        /// </summary>
        private void registerActionItem()
        {
            ICommandRegistry cmd = IoC.Get<ICommandRegistry>();
            IUXConfiguration uxConfig = IoC.Get<IUXConfiguration>();

            var uxSite = new UXSiteSetup
            {
                UXSiteIdPath = "VcTabTeach/VcRibbonTools",
                UXSiteType = UXSiteType.RibbonGroup
            };

            uxConfig.RegisterSite(uxSite);

            IActionItem item = cmd.FindItem("Snap3DGraphics");

            var uxSiteSetup = new UXSiteSetup()
            {
                UXSiteIdPath = "VcTabTeach/VcRibbonTools",
                EntryId = item.Id
            };

            cmd.RegisterActionItem(item, uxSiteSetup);
        }
    }
}
